import typer
import re
import traceback
import numpy as np
from rich import print
from pathlib import Path
from tqdm import tqdm
from dataclasses import dataclass, field
from itertools import combinations, chain
from typing import Optional
from langdetect import detect, DetectorFactory

from dygest import llms, utils, ner_utils, prompts, json_schemas
from dygest.translations import LANGUAGES
from dygest.output_utils import ExportFormats
from dygest.ner_utils import NERlanguages


@dataclass
class DygestBaseParams:
    filepath: str
    output_dir: str = "./output"
    light_model: Optional[str] = None
    expert_model: Optional[str] = None
    embedding_model: Optional[str] = None
    temperature: float = 0.1
    sleep: float = 0
    chunk_size: int = 1000
    add_toc: bool = False
    add_summaries: bool = False
    add_keywords: bool = False
    add_ner: bool = True
    sim_threshold: float = 0.8
    provided_language: NERlanguages = NERlanguages.AUTO
    precise: bool = False
    verbose: bool = False
    export_metadata: bool = False
    export_format: ExportFormats = ExportFormats.HTML


@dataclass
class DygestProcessor(DygestBaseParams):
    ner_tagger: Optional[object] = field(default=None, init=False)
    text: Optional[str] = field(default=None, init=False)
    chunks: Optional[dict] = field(default=None, init=False)
    sentence_offsets: Optional[str] = field(default=None, init=False)
    toc: Optional[list] = field(default=None, init=False)
    summaries: Optional[str] = field(default=None, init=False)
    keywords: Optional[list] = field(default=None, init=False)
    entities: Optional[list] = field(default=None, init=False)
    token_count: Optional[int] = field(default=None, init=False)
    language_ISO: NERlanguages = field(default=None, init=False)
    language_string: str = field(default=None, init=False)
    
    def __post_init__(self):
        self.output_dir = Path(self.output_dir)
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True)
            if self.verbose:
                print(f"... Created output directory at {self.output_dir}") 

    def run_language_detection(self) -> str:
        """
        Get language of text to set the correct NER model.
        
        Returns:
            str: An ISO 639-1 language code ('en', 'de', 'es' ...)
        """
        language = self.provided_language
            
        if language == 'auto':
            DetectorFactory.seed = 0
            language_ISO = detect(self.text[:500])

            if self.verbose:
                print(f"... Detected language '{language_ISO}'")
        else:
            language_ISO = language
        return language_ISO   
        
    def run_ner(self) -> list[str] | None:
        """
        Run Named Entity Recognition with flair framework on the text.
        """
        # Load NER tagger if not already loaded or if language has changed
        if self.ner_tagger is None or self.provided_language == 'auto':
            self.ner_tagger = ner_utils.load_tagger(
                language=self.language_ISO,
                precise=self.precise
            )
            if self.verbose:
                print(f"... Loaded NER tagger for language: {self.language_ISO}")

            # Run Named Entity Recognition (NER)
            entities = ner_utils.get_flair_entities(self.text, self.ner_tagger)
            all_entities = ner_utils.update_entity_positions(entities, self.text)

            if self.verbose:
                print(f"... ENTITIES FOR DOC:")
                utils.print_entities(all_entities)
            
            return all_entities
        else:
            return []
        
    def create_toc(self) -> dict:
        """
        Create a Table of Contents (TOC) for the provided file.
        """
        def is_valid_location(location: str) -> bool:
            """
            Validate if a topic location follows this sentence id ("s_id")
            structure: "S362"
            """
            return bool(re.fullmatch(r'S[1-9]\d*$', location))
        
        def fix_wrong_location(location: str) -> str:
            """
            Try to fix a malformed topic location structure.
            
            Transforms strings like:
            - "S<019>" to "S19"
            - "S09" to "S9"
            """
            if re.fullmatch(r'S<\d+>', location):
                intermediate = re.sub(r'[<>]', '', location)
                fixed = re.sub(r'S0+', 'S', intermediate)
                return None if fixed == 'S' else fixed
            elif re.fullmatch(r'S0+\d+', location):
                fixed = re.sub(r'S0+', 'S', location)
                return None if fixed == 'S' else fixed
            return None

        def align_toc_part(toc_part: list[dict], chunk: dict) -> list[dict]:
            """
            Align the topic locations (e.g. "S7") to match them with the 
            correct chunk sentence IDs. If the location from the TOC is 
            outside the range of the chunk’s sentence IDs, adjust it to 
            the sentence ID within the chunk.
            """
            # Check for a structured json output ('topics' key is present)
            if 'topics' in toc_part:
                toc_part = toc_part['topics']

            # Extract numeric sentence IDs from the chunk
            chunk_start = chunk['s_ids'][0]   # e.g. 'S47'
            chunk_end   = chunk['s_ids'][-1]  # e.g. 'S48'
            start_num   = int(chunk_start[1:])  
            end_num     = int(chunk_end[1:])      

            # Validate locations in toc_part
            all_toc_nums = [int(topic['location'][1:]) for topic in toc_part
                            if is_valid_location(topic['location'])]
            if len(all_toc_nums) == 0:
                # Fallback if none are valid
                toc_start_num = start_num
                toc_end_num = end_num
            else:
                toc_start_num = min(all_toc_nums)
                toc_end_num = max(all_toc_nums)
            
            aligned_toc_part = []

            for topic in toc_part:
                # Validate location
                if not is_valid_location(topic['location']):
                    location = fix_wrong_location(topic['location'])
                    if location is None:
                        location = chunk_start
                    topic['location'] = location
                else: 
                    location = topic['location']  # e.g. 'S2'
                            
                loc_num = int(location[1:])
                
                # If it's already within [start_num, end_num], leave it alone:
                if start_num <= loc_num <= end_num:
                    new_loc_num = loc_num
                else:
                    # If toc_part effectively has only one point (avoid dividing by zero)
                    if toc_start_num == toc_end_num:
                        # Just clamp to the chunk start or end
                        if loc_num < start_num:
                            new_loc_num = start_num
                        else:
                            new_loc_num = end_num
                    else:
                        # Calculate proportion in [toc_start_num, toc_end_num]
                        proportion = (loc_num - toc_start_num) / (toc_end_num - toc_start_num)
                        # Map proportion to [start_num, end_num]
                        mapped_loc_num = start_num + proportion * (end_num - start_num)
                        # Round and clamp
                        mapped_loc_num = round(mapped_loc_num)
                        mapped_loc_num = max(start_num, min(mapped_loc_num, end_num))
                        new_loc_num = mapped_loc_num

                topic['location'] = f"S{new_loc_num}"
                aligned_toc_part.append(topic)
        
            return aligned_toc_part
           
        # TOC processing
        print(f'... Creating TOC with {self.light_model}')
        
        complete_toc_parts = []
        for chunk_key, chunk_data in tqdm(self.chunks.items()):
            result = llms.call_llm(
                prompt=prompts.build_prompt(
                    template=prompts.GET_TOPICS,
                    first_sentence=chunk_data['s_ids'][0],
                    last_sentence=chunk_data['s_ids'][-1],
                    language=self.language_string,
                    chunk=chunk_data['text']
                    ),
                model=self.light_model,
                output_format='json_schema',
                json_schema=json_schemas.GET_TOPICS_JSON,
                temperature=self.temperature,
                sleep_time=self.sleep
                )
            
            # Validate for correct JSON format
            toc_part = utils.validate_LLM_result(result, self.verbose)
            
            # Re-align topic locations
            toc_part = align_toc_part(toc_part, chunk_data)

            # Append toc_part
            complete_toc_parts.extend(toc_part)

            if self.verbose:
                print(f"... TOC PART FOR {chunk_key.upper()}:")
                utils.print_toc_topics(toc_part)

        # Post-Processing: Remove similar summaries
        print(f'... Removing similar TOC entries')
        sp = SummaryProcessor(
            summaries=complete_toc_parts,
            embedding_model=self.embedding_model,
            key='topic',
            threshold=self.sim_threshold,
            verbose=self.verbose
            )
        filtered_toc_parts = sp.get_filtered_summaries()

        # Post-Processing: Create TOC
        print(f'... Compiling TOC with {self.expert_model}')
        toc = llms.call_llm(
            prompt=prompts.build_prompt(
                template=prompts.CREATE_TOC,
                toc_parts=str(filtered_toc_parts),
                language=self.language_string
                ),
            model=self.expert_model,
            output_format='json_schema',
            json_schema=json_schemas.CREATE_TOC_JSON,
            temperature=self.temperature,
            sleep_time=self.sleep
            )
        
        # Validate LLM response
        toc = utils.validate_LLM_result(toc)
        final_toc = utils.validate_toc(toc)
        
        return final_toc

    def create_summaries_and_keywords(self) -> tuple[dict, dict]:
        """
        Create summaries and keywords in one go.
        """
        print(f'... Creating summary and keywords with {self.light_model}')
        
        summaries = []
        keywords = []
        for chunk_key, chunk_data in tqdm(self.chunks.items()):
            result = llms.call_llm(
                prompt=prompts.build_prompt(
                    template=prompts.CREATE_SUMMARY_AND_KEYWORDS,
                    text_chunk=chunk_data['text'],
                    language=self.language_string
                    ),
                model=self.light_model,
                output_format='json_schema',
                json_schema=json_schemas.CREATE_TOC_JSON,
                temperature=self.temperature,
                sleep_time=self.sleep
                )
            
            # Validate
            validated_result = utils.validate_LLM_result(result)
                        
            # Append
            summaries.append(validated_result['summary'])
            keywords.append(validated_result['keywords'])
            
            if self.verbose:
                print(f"... SUMMARY FOR {chunk_key.upper()}:")
                utils.print_summaries(validated_result['summary'])
                print("...")
                print(f"... KEYWORDS FOR {chunk_key.upper()}:")
                utils.print_summaries(validated_result['keywords'])
                
        print(f'... Harmonizing summaries with {self.expert_model}')
        final_summary = llms.call_llm(
            prompt=prompts.build_prompt(
                template=prompts.COMBINE_SUMMARIES,
                summaries='\n'.join(summaries),
                language=self.language_string
                ),
            model=self.expert_model,
            temperature=self.temperature,
            sleep_time=self.sleep 
            )

        # Create a unique set of keywords
        keywords_for_doc = self.clean_generated_keywords(keywords)
        
        # Remove similar keywords
        print(f'... Removing similar keywords')
        sp = SummaryProcessor(
            keywords=keywords_for_doc,
            embedding_model=self.embedding_model,
            key='topic',
            threshold=self.sim_threshold,
            verbose=self.verbose
            )
        filtered_keywords = sp.get_filtered_keywords()
        
        return final_summary, filtered_keywords
        
    def create_summaries(self) -> str:
        """
        Create summaries.
        """
        print(f'... Creating summary with {self.light_model}')
        
        summaries = []
        for chunk_key, chunk_data in tqdm(self.chunks.items()):
            summary = llms.call_llm(
                prompt=prompts.build_prompt(
                    template=prompts.CREATE_SUMMARY,
                    text_chunk=chunk_data['text'],
                    language=self.language_string
                    ),
                model=self.light_model,
                temperature=self.temperature,
                sleep_time=self.sleep
                )
            summaries.append(summary)
            
            if self.verbose:
                print(f"... SUMMARY FOR {chunk_key.upper()}:")
                utils.print_summaries(summary)
                
        print(f'... Harmonizing summaries with {self.expert_model}')
        final_summary = llms.call_llm(
            prompt=prompts.build_prompt(
                template=prompts.COMBINE_SUMMARIES,
                summaries='\n'.join(summaries),
                language=self.language_string
                ),
            model=self.expert_model,
            temperature=self.temperature,
            sleep_time=self.sleep
            )
        
        return final_summary
    
    def generate_keywords(self):
        """
        Generate keywords for the input text.
        """
        print(f'... Generating keywords with {self.light_model}')
        
        keywords_for_doc = []
        for chunk_key, chunk_data in tqdm(self.chunks.items()):
            keywords_for_chunk = llms.call_llm(
                prompt=prompts.build_prompt(
                    template=prompts.CREATE_KEYWORDS,
                    text_chunk=chunk_data['text'],
                    language=self.language_string
                    ),
                model=self.light_model,
                temperature=self.temperature,
                sleep_time=self.sleep
                )
            keywords_for_doc.append(keywords_for_chunk.split(','))
            
            if self.verbose:
                print(f"... KEYWORDS FOR CHUNK {chunk_key.upper()}:")
                utils.print_summaries(keywords_for_chunk)

        # Create a unique set of keywords
        keywords_for_doc = self.clean_generated_keywords(keywords_for_doc)

        # Remove similar keywords
        print(f'... Removing similar keywords')
        sp = SummaryProcessor(
            keywords=keywords_for_doc,
            embedding_model=self.embedding_model,
            key='topic',
            threshold=self.sim_threshold,
            verbose=self.verbose
            )
        filtered_keywords = sp.get_filtered_keywords()

        return filtered_keywords
    
    def clean_generated_keywords(
        self, 
        keywords_for_doc: str | list
        ) -> list[str]:
        """
        Return a unique list of keywords from LLM generated keyword list.
        """        
        clean_keywords = set()
        
        flattened = list(chain.from_iterable(keywords_for_doc))
        for keyword in flattened:
            keyword = utils.remove_punctuation(keyword.strip())
            keyword = utils.replace_underscores_with_whitespace(keyword)
            clean_keywords.add(keyword)

        return clean_keywords
    
    def reset_processing_vals(self):
        """
        Reset processing values for each file.
        """
        self.text = None
        self.chunks = None
        self.toc = None
        self.summaries = None
        self.keywords = None
        self.entities = None
        self.token_count = None
        self.language_ISO = None
        self.language_string = None
        self.sentence_offsets = None

    def process_file(self, file: Path):
        """
        Main function for processing files and creating TOCs, summaries, 
        keywords as well as running NER.
        """
        # Reset processing values
        self.reset_processing_vals()

        # Get filename and output filepath
        self.filename = file.stem
        self.output_filepath = self.output_dir.joinpath(self.filename)

        print(f'[bold][orange]... Dygesting {file.name}')

        # Load file
        self.text = self.run_with_error_handling(
            utils.load_file,
            file,
            self.verbose,
            error_message="Error during file loading"
        )
        
        # Chunk file
        self.chunks, self.token_count, self.sentence_offsets = (
            self.run_with_error_handling(
                utils.chunk_text, 
                text=self.text, 
                chunk_size=self.chunk_size, 
                error_message="Error during text chunking"
                )
            )

        if self.verbose:
            print(f"... Total tokens in file: {self.token_count}")
            print(f"... Number of chunks: {len(self.chunks)}")
            
        # Run language detection 
        if not self.language_ISO:
            self.language_ISO = self.run_with_error_handling(
                self.run_language_detection,
                error_message="Error during language detection"
            )
            
        # Transform ISO code to string ('en' → 'English')
        self.language_string = LANGUAGES.get(self.language_ISO).title()
        
        # Run Named Entity Recognition (NER)
        if self.add_ner:
            self.entities = self.run_with_error_handling(
                self.run_ner,
                error_message="Error during NER task"
            )
        
        # Create TOC (Table of Contents)
        if self.add_toc:
            self.toc = self.run_with_error_handling(
                self.create_toc,
                error_message="Error during TOC creation"
            )
         
        # Create summary and keywords in one go       
        if self.add_summaries and self.add_keywords:
            self.summaries, self.keywords = self.run_with_error_handling(
                self.create_summaries_and_keywords, 
                error_message="Error during creating summaries and keywords"
            )

        # Create only summary
        elif self.add_summaries:
            
            self.summaries = self.run_with_error_handling(
                self.create_summaries, 
                error_message="Error during summary creation"
            )

        # Create only keywords
        elif self.add_keywords:
            self.keywords = self.run_with_error_handling(
                self.generate_keywords,
                error_message="Error during keyword generation"
            )

    def run_with_error_handling(
        self, 
        func, 
        *args, 
        error_message="", 
        **kwargs
        ):
        """
        Helper function to handle exceptions uniformly.
        """
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(f"[purple]... {error_message}: {e}")
            print(f"[purple]{traceback.format_exc()}")
            raise typer.Exit(code=1)
            
            
class SummaryProcessor:
    def __init__(
            self, 
            summaries: Optional[dict[str]] = None, 
            keywords: Optional[set[str]] = None,
            embedding_model: str = None,
            key: str = 'topic', 
            threshold: float = 0.8,
            verbose: bool = False
            ):
        """
        Initialize the SummaryProcessor and process the summaries.

        Args:
            summaries (list[dict]): List of summary dictionaries.
            embedding_model (str): The name of the embedding model to use.
            key (str): The key in the dictionary to embed (e.g., 'topic').
            threshold (float): Cosine similarity threshold to consider topics as similar.
        """
        self.embedding_model = embedding_model
        self.key = key
        self.threshold = threshold
        self.summaries = summaries
        self.keywords = keywords
        self.verbose = verbose  

        self.embedded_summaries = {}
        self.filtered_summaries = []  
        self.embedded_keywords = {}
        self.filtered_keywords = []

    def embed_summaries(self):
        """
        Embed summaries.
        """
        for summary in self.summaries:
            text = summary[self.key]
            response = llms.get_embeddings(text, model=self.embedding_model)
            self.embedded_summaries[text] = np.array(response)

    def embed_keywords(self):
        """
        Embed strings (like keywords).
        """
        for keyword in self.keywords:
            response = llms.get_embeddings(keyword, model=self.embedding_model)
            self.embedded_keywords[keyword] = np.array(response)

    @staticmethod
    def cosine_similarity(vec1, vec2):
        """
        Calculate the cosine similarity between two vectors.

        Args:
            vec1 (np.array): First embedding vector.
            vec2 (np.array): Second embedding vector.

        Returns:
            float: Cosine similarity score.
        """
        dot_product = np.dot(vec1, vec2)
        norm_vec1 = np.linalg.norm(vec1)
        norm_vec2 = np.linalg.norm(vec2)
        if norm_vec1 == 0 or norm_vec2 == 0:
            return 0.0
        return dot_product / (norm_vec1 * norm_vec2)

    def remove_similar_summaries(self):
        """
        Remove similar summaries based on cosine similarity of their topics.

        Returns:
            list[dict]: Filtered list of summaries with similar topics removed.
        """
        similar_topics = []

        # Generate all unique pairs of topics
        for (topic1, emb1), (topic2, emb2) in combinations(self.embedded_summaries.items(), 2):
            similarity = self.cosine_similarity(emb1, emb2)
            if similarity >= self.threshold:
                similar_topics.append({
                    'topic_1': topic1,
                    'topic_2': topic2,
                    'similarity_score': similarity
                })

        # Identify Topics to Remove (topic_2 in each similar pair)
        topics_to_remove = set(pair['topic_2'] for pair in similar_topics)

        # Filter Out Summaries with Topics to Remove
        filtered_summaries = [summary for summary in self.summaries if summary[self.key] not in topics_to_remove]

        # Display Similar Topics Identified
        if self.verbose:
            if similar_topics:
                print("... Similar Topics Identified:")
                for pair in similar_topics:
                    print(f"... '{pair['topic_1']}' <--> '{pair['topic_2']}' with similarity score of {pair['similarity_score']:.4f}")
            else:
                print("... No similar topics found above the threshold.")

        return filtered_summaries
    
    def remove_similar_keywords(self):
        """
        Remove similar keywords based on cosine similarity of their embeddings.

        Returns:
            list[str]: A filtered list of keywords with similar ones removed.
        """
        similar_pairs = []

        # Generate all unique pairs of keywords and check their similarity
        for (kw1, emb1), (kw2, emb2) in combinations(self.embedded_keywords.items(), 2):
            similarity = self.cosine_similarity(emb1, emb2)
            if similarity >= self.threshold:
                similar_pairs.append({
                    'keyword_1': kw1,
                    'keyword_2': kw2,
                    'similarity_score': similarity
                })

        # Identify keywords to remove.
        keywords_to_remove = set(pair['keyword_2'] for pair in similar_pairs)

        # Filter out the similar keywords
        filtered_keywords = [k for k in self.keywords if k not in keywords_to_remove]

        # Display similar keywords if verbose is enabled
        if self.verbose:
            if similar_pairs:
                print("... Similar Keywords Identified:")
                for pair in similar_pairs:
                    print(f"... '{pair['keyword_1']}' <--> '{pair['keyword_2']}' with similarity score of {pair['similarity_score']:.4f}")
            else:
                print("... No similar keywords found above the threshold.")

        return filtered_keywords
    
    def get_filtered_summaries(self):
        """
        Get the filtered summaries.

        Returns:
            list[dict]: Filtered summaries.
        """
        self.embed_summaries()
        self.filtered_summaries = self.remove_similar_summaries()
        return self.filtered_summaries
    
    def get_filtered_keywords(self):
        """
        Get the filtered summaries.

        Returns:
            list[dict]: Filtered summaries.
        """
        self.embed_keywords()
        self.filtered_keywords = self.remove_similar_keywords()
        return self.filtered_keywords


import re
import copy
import time
from enum import Enum
from collections import defaultdict


class NERlanguages(str, Enum):
    AUTO = 'auto'
    AR = 'ar'
    DE = 'de'
    DA = 'da'
    EN = 'en'
    FR = 'fr'
    ES = 'es'
    NL = 'nl'


def load_tagger(language: str = 'en', precise: bool = False):
    """
    Loads flair Classifier depending on a provided language.
    """
    from flair.nn import Classifier
    flair_models = {
        'en': 'ner',
        'ar': 'ar-ner',
        'de': 'de-ner',
        'da': 'da-ner',
        'fr': 'fr-ner',
        'es': 'es-ner',
        'nl': 'nl-ner',
    }
    
    start_time = time.time()
    print('... Loading NER tagger')

    # Load Classifier
    try:
        if precise and language in flair_models:
            tagger = Classifier.load(f'{flair_models[language]}-large')
        elif language not in flair_models:
            tagger = Classifier.load('ner-large')
        else:
            tagger = Classifier.load(f'{flair_models[language]}')
    except {} as e:
        print(f'ERROR: {e}')

    end_time = time.time()
    print(f'... Finished loading NER tagger in {end_time - start_time:.2f} seconds.')

    return tagger

def get_flair_entities(text: str, tagger) -> list:
    """
    Tags named entities using a flair Classifier.
    """
    from flair.splitter import SegtokSentenceSplitter
    
    # Split text_chunk into sentences
    splitter = SegtokSentenceSplitter()
    sentences = splitter.split(text)

    # Predict entities
    print(f'... Starting NER task')
    tagger.predict(sentences, verbose=True)

    entities = []
    for sentence in sentences:
        for entity in sentence.get_spans():
            entities.append({
                'start': entity.start_position,
                'end': entity.end_position,
                'ner_tag': entity.tag,
                'text': entity.text
                })

    return entities

def update_entity_positions(
        entities: list[dict], 
        text_doc: str, 
        verbose: bool = False
        ) -> list[dict]:
    """
    Updates the 'start' and 'end' positions of each entity based on their
    occurrences in the document.

    Parameters:
    - entities (list of dict): The list of entity dictionaries to update.
    - doc (str): The document string where entities are located.

    Returns:
    - list of dict: The updated list of entity dictionaries.
    """
    temp_entities = copy.deepcopy(entities)

    # Group entities by their 'text' value
    text_to_entities = defaultdict(list)
    for entity in temp_entities:
        text_to_entities[entity['text']].append(entity)
    
    # Iterate through each unique 'text' and find all matches in 'doc'
    for text, entity_list in text_to_entities.items():
        # Escape the 'text' to handle any special regex characters
        escaped_text = re.escape(text)
        matches = list(re.finditer(escaped_text, text_doc))
        
        # Count the number of matches and entities
        num_matches = len(matches)
        num_entities = len(entity_list)
        
        if verbose:
            print(
                f"... Processing '{text}': {num_matches} matches found, " 
                f"... {num_entities} entities to update."
                )
            
            # Check for mismatches in counts
            if num_matches < num_entities:
                print(
                    f"... Warning: Not enough matches for '{text}'. "
                    f"... {num_entities - num_matches} entities will not be updated."
                    )
            elif num_matches > num_entities:
                print(
                    f"... Warning: More matches than entities for '{text}'. "
                    f"... {num_matches - num_entities} extra matches will be ignored."
                    )
            
        # Assign match positions to entities
        for i, entity in enumerate(entity_list):
            if i < num_matches:
                match = matches[i]
                entity['start'] = match.start()
                entity['end'] = match.end()
                if verbose:
                    print(
                        f"... Updated entity {i+1}: start={match.start()}, "
                        f"... end={match.end()}"
                        )
            else:
                # Handle entities without corresponding matches
                if verbose:
                    print(
                        f"... Entity {i+1} with text '{text}' has no corresponding "
                        f"... match in 'doc'."
                        )
    
    return temp_entities



### Function for building prompts with kwargs ###

def build_prompt(template, **kwargs):
    prompt = template
    for key, value in kwargs.items():
        placeholder = '{' + key + '}'
        prompt = prompt.replace(placeholder, value)
    return prompt


### TOC CREATION ###

GET_TOPICS =  """You will be given a text composed of multiple sentences. The sentences are numbered sequentially:
- {first_sentence} = the first sentence of the text
- {last_sentence} = the last sentence of the text

Do not include these sentence numbers in your summary. Instead, reference the location **with the sentence number: "S<number>"**.

Your tasks:
1. Identify the top 2 most important topics discussed in the text.
  - Each topic name should be a concise phrase (no more than 5 words).
2. For each topic, generate a precise sub-headline (no more than 8 words) with key details.
3. Count the sentences to give a **precise sentence number** as location for the identified topics.
4. Return results as JSON **IN {language}**. Do NOT add anything else and use this template without any alterations:
  
[
  {
    "topic": "Concise Subheading in {language}",
    "summary": "Concise sub-headline in {language}",
    "location": "S<number>"
  },
  ...
]

**Input Text:**
{chunk}
"""

CREATE_TOC = """
Perform the following tasks on the provided list of summaries and **ALWAYS match their language**:

1. **Create a Table of Contents (TOC):**
   - Group related summaries under concise, relevant headlines.
   - Each headline should encapsulate the theme of its grouped topics.

2. **Maintain Summary Locations:**
   - Do **not** alter the "location" value of any summary.

**Output Requirements:**
- **Format:** JSON
- **Language:** {language}
- **Strictly adhere to the following template and return nothing else:**

[
  {
    "headline": "Headline Name",
    "topics": [
      {
        "summary": "summary",
        "location": "location"
      }
    ]
  },
]

**Input Summaries:**
{toc_parts}
"""

### SUMMARY CREATION ###

CREATE_SUMMARY_AND_KEYWORDS = """
Perform the following tasks on the provided text:

1. Generate a summary **IN {language}**:
  - max. 3 sentences with the **most important topics** for the provided text.
2. Incorporate key details in your summary:
  - **Who**: Names of individuals involved.
  - **What**: Description of the events or actions.
  - **When**: Mentioned dates.
  - **Where**: Locations related to the events.
  - **Why**: Purpose or reason behind the events.
3. Generate a **set of broad, high-level keywords** that capture the main topics of the provided text chunk.
  - Focus on **overarching themes and main topics** rather than specific details.
  - Ensure keywords are **accurate** and **informative**.
  - Limit the number of keywords to **2-5** to maintain relevance.

**Output Requirements:**
- **Format:** JSON
- **Language:** {language}
- **Strictly adhere to the following template and return nothing else:**

{
  "summary": "Summary",
  "keywords": ["keyword1", "keyword2", ...]
}

**Text Chunk:**
{text_chunk}
"""

CREATE_SUMMARY = """
Perform the following tasks on the provided text chunk:

1. Generate a **summary of max. 3 sentences** with the **most important topics** for the provided text chunk.
2. Incorporate key details in your summary:
   - **Who**: Names of individuals involved.
   - **What**: Description of the events or actions.
   - **When**: Mentioned dates.
   - **Where**: Locations related to the events.
   - **Why**: Purpose or reason behind the events.
3. Return the results in **{language}**.

**Return ONLY the summary and add nothing else!**

**Text Chunk:**
{text_chunk}
"""

COMBINE_SUMMARIES = """
Perform the following tasks on the provided list of summaries:

1. Generate a **single summary of maximum 5 sentences** that captures the **most important topics** from all provided summaries.
2. **Remove similar summaries** to ensure the remaining summaries are **unique**.
3. Return the summary in **{language}** as a **plain string**.

Return **ONLY** the single summary.

**Input Summaries:**
{summaries}
"""

### KEYWORD CREATION ###

CREATE_KEYWORDS = """
Analyze the following text chunk and generate a set of broad, high-level keywords that capture the main topics and themes of the content.

**Requirements:**
1. **Focus on overarching themes and main topics** rather than specific details.
2. Ensure keywords are **accurate** and **informative**.
3. Provide keywords **{language}** as the input text.
4. Limit the number of keywords to **2-5** to maintain relevance.
5. Return the keywords as a plain string, separated by commas.

Return **ONLY the comma-separated keywords** and **nothing else**.

**Text Chunk:**
{text_chunk}
"""


import json
import re
import json_repair
import typer

from rich import print
from typing import Optional
from pathlib import Path


VALID_FILE_FORMATS = [
    '.txt',
    '.csv',
    '.xlsx',
    '.doc',
    '.docx',
    '.pdf',
    '.html',
    '.xml'
]


def resolve_input_dir(filepath: str = None, output_dir: str = None) -> Path:
    """
    Resolve a filepath (file or folder).
    """
    input_path = Path(filepath)

    if not input_path.exists():
        print(f"... Error: The path '{filepath}' does not exist.")
        return

    if output_dir is None:
        if input_path.is_dir():
            resolved_output_dir = input_path
        else:
            resolved_output_dir = input_path.parent
    else:
        resolved_output_dir = Path(output_dir)
        
    return resolved_output_dir

def load_filepath(filepath: str) -> list[Path]:
    """
    Loads a filepath (file or folder).
    """
    filepath = Path(filepath)
    if not filepath.exists():
        print("... Please provide a valid filepath.")
        return
    elif filepath.is_dir():
        files_to_process = []        
        for fpath in list(filepath.rglob('*.*')):
            if Path(fpath).is_file() and Path(fpath).suffix in VALID_FILE_FORMATS:
                normalized_fpath = normalize_filepath(fpath)
                files_to_process.append(normalized_fpath)
            else:
                print(f"... Skipping [bold]{fpath}[/bold] as the file format is not supported.")
                
        if not files_to_process:
            print("... No files with valid file formats found in the directory.")
            raise typer.Exit(code=1)
    else:
        files_to_process = [filepath]
        
    return files_to_process

def normalize_filepath(filepath: Path) -> Path:
    """
    Normalizes the filepath by replacing non-word characters with underscores,
    collapsing multiple underscores into one, and removing leading/trailing underscores.
    """
    new_filename = re.sub(r'\W+', '_', filepath.stem)
    new_filename = new_filename.strip('_')
    new_filename = re.sub(r'_+', '_', new_filename)
    
    suffix = filepath.suffix.lower()
    
    # Construct the new path
    new_path = filepath.parent / f"{new_filename}{suffix}"
    
    # Rename the file
    filepath.rename(new_path)

    return new_path.resolve()

def dehyphenate(text: str) -> str:
    """
    Remove hyphenation caused by line breaks.
    """
    return re.sub(r'-\s*\n\s*', '', text)

def remove_linebreaks(text: str) -> str:
    """
    Remove linebreaks that cut sentences.
    """
    return re.sub(r"([A-Za-z0-9\.,:;!“”'\?])\n([A-Za-z0-9\.,:;!“”'\?])", r'\1 \2', text)

def load_file(fpath: str, verbose: bool = False) -> str:
    """
    Load file and convert it to Markdown, then clean the text.
    
    Parameters:
    - fpath: Path to the input file.

    TODO: Better handling of PDF layout issues (redundant headers, footers etc.)
    """
    from markitdown import MarkItDown
    
    if verbose:
        print(f"... Converting to Markdown")

    # Convert to markdown
    md = MarkItDown()
    result = md.convert(str(fpath))
    
    # Clean multiple whitespaces
    raw_text = re.sub(r' {2,}', ' ', result.text_content)
        
    # Check if the file is a PDF
    if Path(fpath).suffix.lower() == '.pdf':
        if verbose:
            print("... Performing PDF-specific cleaning.")
        
        raw_text = dehyphenate(raw_text)
        raw_text = remove_linebreaks(raw_text)

        if verbose:
            print("... Dehyphenation completed.")
            print("... Linebreaks removed.")
    
    return raw_text
    
def remove_punctuation(text: str) -> str:
    return re.sub(r'[^\w\s]', '', text)

def replace_underscores_with_whitespace(text: str) -> str:
    return re.sub(r'_', ' ', text)

def chunk_text(
    text: str, 
    chunk_size: int = 1000
) -> tuple[dict, int, dict]:
    """
    Chunks the text sentence-wise by max_tokens.
    
    Returns:
    - chunks: A dictionary of the form:
        {
          'chunk_01': {
              'text': '...',
              's_ids': ['S1', 'S2', ...]
          },
          'chunk_02': {...},
          ...
        }
    - token_count: Total number of tokens in the input text.
    - sentence_offsets: A dict mapping "S<id>" to the starting character 
      index of that sentence in 'text'.
    """
    import tiktoken
    from flair.splitter import SegtokSentenceSplitter

    tokenizer = tiktoken.get_encoding("gpt2")

    token_count = 0
    chunks = {}
    chunk_index = 1
    current_chunk = []
    current_chunk_length = 0
    current_chunk_s_ids = []

    # Remove all linebreaks from the input text
    text = re.sub(r'[\n\t\r]', ' ', text)

    # Split text into sentences
    splitter = SegtokSentenceSplitter()
    sentences = splitter.split(text)

    # Map sentence IDs to their start offsets in the input text
    sentence_offsets = {}
    current_search_pos = 0
    sentence_id = 1

    for sentence in sentences:
        # Find the offset of this sentence in the original text
        start_index = text.index(sentence.text[:30], current_search_pos)
        sentence_offsets[f"S{sentence_id}"] = start_index
        current_search_pos = start_index + len(sentence.text)
        sentence_id += 1

    # Chunk text based on token limits
    current_chunk_s_ids = []
    global_sentence_id = 1  # reset to 1 to track sentences globally

    for sentence in sentences:
        s_id = f"S{global_sentence_id}"
        sentence_tokens = tokenizer.encode(sentence.text)
        sentence_length = len(sentence_tokens)
        token_count += sentence_length

        if current_chunk_length + sentence_length > chunk_size:
            # Finish the current chunk
            if current_chunk:
                chunk_key = f"chunk_{chunk_index:02d}"
                chunks[chunk_key] = {
                    'text': tokenizer.decode(current_chunk),
                    's_ids': current_chunk_s_ids
                }
                chunk_index += 1
            
            # Start a new chunk
            current_chunk = sentence_tokens[:]
            current_chunk_length = sentence_length
            current_chunk_s_ids = [s_id]
        else:
            # Add sentence to the current chunk
            if current_chunk:  # Add a space between sentences
                space_token = tokenizer.encode(" ")[0]
                current_chunk.append(space_token)
                current_chunk_length += 1
            current_chunk.extend(sentence_tokens)
            current_chunk_length += sentence_length
            current_chunk_s_ids.append(s_id)

        global_sentence_id += 1

    # Add the final chunk if present
    if current_chunk:
        chunk_key = f"chunk_{chunk_index:02d}"
        chunks[chunk_key] = {
            'text': tokenizer.decode(current_chunk),
            's_ids': current_chunk_s_ids
        }

    return chunks, token_count, sentence_offsets

def sort_summaries_by_key(summaries: list[dict], key: str) -> list[dict]:
    return sorted(summaries, key=lambda x: x[key])

def chunk_summaries(
    summaries: list[dict], 
    limit: int = 15, 
    sort_by_key: str = None
    ) -> list[list[dict]]:
    if sort_by_key:
        summaries = sort_summaries_by_key(summaries, sort_by_key)

    chunked_summaries = []
    for i in range(0, len(summaries), limit):
        chunk = summaries[i:i + limit]
        chunked_summaries.append(chunk)
    return chunked_summaries

def is_valid_json(json_string: str, verbose: bool = False) -> bool:
    """
    Check if json_string is valid JSON.
    """
    try:
        json.loads(json_string)
        return True
    except json.JSONDecodeError as e:
        if verbose:
            print(f"... Invalid JSON: {e}")
        return False

def fix_json(json_string: str) -> str:
    """
    Cleans and repairs a string to return valid JSON.
    """
    # Remove trailing Markdown code block markers (```json and ```)
    json_string = re.sub(r"```json\s*", '', json_string, flags=re.IGNORECASE)
    json_string = re.sub(r"```\s*", '', json_string)
    
    # Repair other json errors
    json_string = json_repair.repair_json(json_string)

    return json_string

def validate_and_fix_json(
        json_string: str, 
        verbose: bool = False
    ) -> Optional[dict]:
    """
    Check if a string is valid JSON. If not attempt to make it valid.
    """
    # Check if string is valid json string
    if is_valid_json(json_string, verbose):
        return json.loads(json_string)
    
    if verbose:
        print("... Attempting to fix JSON.")

    # Attempt to fix json
    fixed_json_str = fix_json(json_string)
    
    if is_valid_json(fixed_json_str):
        if verbose:
            print("... JSON fixed successfully.")
        return json.loads(fixed_json_str)
    else:
        print("... Failed to fix JSON.")
        print(f'\n{json_string}\n')
        return None

def validate_LLM_result(llm_result: str, verbose: bool = False):
    """
    Validate a collection of LLM summaries for correct JSON format.
    """
    try:
        result = validate_and_fix_json(llm_result, verbose)
    except json.JSONDecodeError:
        print("... Error decoding JSON.\n")
        print(llm_result)
    return result

def validate_toc(toc: dict) -> list:
    """
    Extract the 'properties' dictionary from a TOC dict, if present.

    This function checks if the input dictionary `toc` contains a key named
    'properties'. If the key exists, it returns the corresponding list. If 
    the 'properties' key is not found, it returns the original `toc` list.
    """
    if 'properties' in toc:
        return toc['properties']
    elif 'items' in toc:
        return toc['items']
    else:
        return [toc] if not isinstance(toc, list) else toc

def print_entities(entities: list[dict]) -> None:
    [print(f"... ... {entity['text']} → {entity['ner_tag']}") for entity in entities]

def print_toc_topics(toc_part: list[dict]) -> None:
    for idx, item in enumerate(toc_part):
        print(f"... [{idx + 1}] {item['topic']}: {item['summary']} (LOCATION: {item['location']})")

def print_summaries(summary: str) -> None:
    print(f"... {summary}")
    


import typer
from typing import Optional
from pathlib import Path
from rich import print
from dygest.output_utils import ExportFormats
from dygest.config import CONFIG, DEFAULT_CONFIG

app = typer.Typer(
    no_args_is_help=True,
    help='🌞 DYGEST: Document Insights Generator 🌞'
)

@app.command("config", no_args_is_help=True)
def configure(
    light_model: str = typer.Option(
        None,
        "--light_model",
        "-l",
        help="LLM model name for lighter tasks (summarization, keywords)",
    ),
    expert_model: str = typer.Option(
        None,
        "--expert_model",
        "-x",
        help="LLM model name for heavier tasks (TOCs).",
    ),
    embedding_model: str = typer.Option(
        None,
        "--embedding_model",
        "-e",
        help="Embedding model name.",
    ),
    temperature: float = typer.Option(
        None,
        "--temperature",
        "-t",
        help='Temperature of LLM.',
    ),
    sleep: float = typer.Option(
        None,
        "--sleep",
        "-s",
        help='Pause LLM requests to prevent rate limit errors (in seconds).',
    ),
    chunk_size: int = typer.Option(
        None,
        "--chunk_size",
        "-c",
        help="Maximum number of tokens per chunk."
    ),
    ner: Optional[bool] = typer.Option(
        None,
        "--ner/--no-ner",
        help="Enable Named Entity Recognition (NER). Defaults to False.",
    ),
    precise: Optional[bool] = typer.Option(
        None,
        "--precise/--fast",
        help="Enable precise mode for NER. Defaults to fast mode.",
    ),
    language: str = typer.Option(
        None,
        "--lang",
        "-lang",
        help='Language of file(s) for NER. Defaults to auto-detection.',
    ),
    api_base: str = typer.Option(
        None,
        "--api_base",
        '-api',
        help="Set custom API base url for providers like Ollama and Hugginface."
    ),
    view_config: bool = typer.Option(
        False,
        "--view_config",
        "-v",
        help="View loaded config parameters.",
    )
):
    """
    Configure LLMs, Embeddings and Named Entity Recognition.
    """
    from dygest.config import load_config, save_config, print_config
    from dygest.ner_utils import NERlanguages
    
    global CONFIG
    if CONFIG is None:
        CONFIG = load_config()
    
    if view_config:
        print_config(CONFIG)
        return
        
    CONFIG['light_model'] = light_model if light_model is not None else CONFIG.get('light_model')
    CONFIG['expert_model'] = expert_model if expert_model is not None else CONFIG.get('expert_model')
    CONFIG['embedding_model'] = embedding_model if embedding_model is not None else CONFIG.get('embedding_model')
    CONFIG['temperature'] = temperature if temperature is not None else CONFIG.get('temperature')
    CONFIG['sleep'] = sleep if sleep is not None else CONFIG.get('sleep')
    CONFIG['api_base'] = api_base if api_base is not None else CONFIG.get('api_base')
    CONFIG['chunk_size'] = chunk_size if chunk_size is not None else CONFIG.get('chunk_size')
    CONFIG['ner'] = ner if ner is not None else CONFIG.get('ner')
    CONFIG['precise'] = precise if precise is not None else CONFIG.get('precise')

    # Convert language if provided
    if language is not None:
        try:
            CONFIG['language'] = NERlanguages(language)
        except ValueError:
            typer.echo(f"... Warning: '{language}' is not a valid NER language. Using 'auto' instead.", err=True)
            CONFIG['language'] = NERlanguages.AUTO
    
    # Save and print config
    save_config(CONFIG)
    print_config(CONFIG)

@app.command("run", no_args_is_help=True)
def main(
    filepath: str = typer.Option(
        None,
        "--files",
        "-f",
        help="Path to the input folder or .txt file."
    ),
    output_dir: Optional[str] = typer.Option(
        None,
        "--output_dir",
        "-o",
        help="If not provided, outputs will be saved in the input folder.",
    ),
    export_format: Optional[ExportFormats] = typer.Option(
        ExportFormats.HTML,
        "--export_format",
        "-ex",
        help="Set the data format for exporting.",
    ),
    toc: bool = typer.Option(
        False,
        "--toc",
        "-t",
        help="Create a Table of Contents (TOC) for the text. Defaults to False.",
    ),
    summarize: bool = typer.Option(
        False,
        "--summarize",
        "-s",
        help="Include a short summary for the text. Defaults to False.",
    ),
    keywords: bool = typer.Option(
        False,
        "--keywords",
        "-k",
        help="Create descriptive keywords for the text. Defaults to False.",
    ),
    sim_threshold: float = typer.Option(
        0.85,
        "--sim_threshold",
        "-sim",
        help="Similarity threshold for removing duplicate topics."
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Enable verbose output. Defaults to False.",
    ),
    export_metadata: bool = typer.Option(
        False,
        "--export_metadata",
        "-meta",
        help="Enable exporting metadata to output file(s). Defaults to False.",
    )
):
    """
    Create insights for your documents (summaries, keywords, TOCs).
    """
    from dygest.config import load_config
    from dygest import core, output_utils, utils
    global CONFIG
    
    # Load config and make sure it is correctly set
    CONFIG = load_config()
    if CONFIG == DEFAULT_CONFIG:
        print(f"[purple]... Please configure dygest first by running *dygest \
config* and set your LLMs.")
        return
    
    # Check for file
    if filepath is None:
        print('... Please provide a file or folder (--files, -f).')
        return
    
    # Create a list of all files to process
    files_to_process = utils.load_filepath(filepath)
        
    # for file in proc.files_to_process:
    for file in files_to_process:
        proc = core.DygestProcessor(
            filepath=filepath,
            output_dir=utils.resolve_input_dir(Path(filepath), output_dir),
            light_model=CONFIG['light_model'],
            expert_model=CONFIG['expert_model'],
            embedding_model=CONFIG['embedding_model'],
            temperature=CONFIG['temperature'],
            sleep=CONFIG['sleep'],
            chunk_size=CONFIG['chunk_size'],
            add_toc=toc,
            add_summaries=summarize,
            add_keywords=keywords,
            add_ner=CONFIG['ner'],
            sim_threshold=sim_threshold,
            provided_language=CONFIG['language'],
            precise=CONFIG['precise'],
            verbose=verbose,
            export_metadata=export_metadata,
            export_format=export_format
        )
        
        # Process file
        proc.process_file(file)

        # Write output
        try:
            formats_to_export = (
                [ExportFormats.CSV, ExportFormats.JSON, ExportFormats.HTML]
                if proc.export_format == ExportFormats.ALL
                else [proc.export_format]
            )
            
            for format in formats_to_export:
                proc.export_format = format
                writer = output_utils.get_writer(proc)
                write_method = getattr(writer, 'write', None)
                write_method()
                print('[blue][bold]... DONE')

        except ValueError as ve:
            print(f'... {ve}')
        except Exception as e:
            print(f'... An unexpected error occurred: {e}')

if __name__ == '__main__':
    app()
    


HTML_CONTENT = """
<html>
<head>
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/normalize/8.0.1/normalize.min.css">
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/skeleton/2.0.4/skeleton.min.css">
<style>
    body {
        background-color: #ffffff;
        font-family: Arial, sans-serif;
        color: #000000;
        font-size: 18px; 
        margin: 20px;
    }

    a {
        color: #3d64bb; 
        text-decoration: none;
    }

    a:hover {
        color: #296af6; 
        text-decoration: underline; 
    }

    .anchor {
        text-decoration: underline;
    }

    button {
        background-color: #03a9f4;
        font-weight: 100;
        color: white;
        border: none;
        padding: 10px 10px;
        text-align: center;
        text-decoration: none;
        display: inline-block;
        font-size: 15px;
        margin: 4px 2px;
        cursor: pointer;
        border-radius: 20px;
        transition: 0.1s;
        text-transform: none;
        line-height: 1;
        vertical-align: middle;
        border-bottom: 2px solid #0074aa;
    }

    button:hover {
        background-color: #0289c7;
        color: white;
        border-top: 2px solid #0078b0;
        border-bottom: 2px solid #0078b0;
    }

    button:active {
        background-color: #03a9f4;
        color: white;
    }

    button:focus {
        background-color: #03a9f4;
        color: white;
    }

    button.save {
        background-color: #1bc89a;
        border-bottom: 2px solid #14916f;
    }

    button.save:hover {
        background-color: #15a17b;
        color: white;
        border-top: 2px solid #117b5f;
        border-bottom: 2px solid #117b5f;
    }

    #back-to-top {
        position: fixed;
        bottom: 20px;
        right: 20px;
        display: none; 
        z-index: 99; 
        background-color: #03a9f4;
        font-weight: 100;
        color: white;
        border: none;
        padding: 10px 10px;
        text-align: center;
        text-decoration: none;
        font-size: 15px;
        cursor: pointer;
        border-radius: 20px;
        transition: 0.1s;
        text-transform: none;
        line-height: 1;
        vertical-align: middle;
        border-bottom: 2px solid #0074aa;
    }

    #back-to-top:hover {
        background-color: #0289c7;
        border-top: 2px solid #0078b0;
        border-bottom: 2px solid #0078b0;
    }

    .content {
        line-height: 1.75;
    }

    .controls {
        padding: 10px;
        border-radius: 10px;
        margin-bottom: 10px;
        text-align: center;
    }

    div[contenteditable="true"] {
        border-radius: 10px;
        padding: 10px;
        min-height: 100px; 
    }

    div[contenteditable="true"]:focus {
        background-color: #f6f6f6;
        border-radius: 10px;
        padding: 10px;
        min-height: 100px; 
    }
    
    .edits {
        border: 0;
    }

    .edits:focus {
        border: 3px solid deeppink;
        background-color: rgb(255, 215, 215);
        padding: 4px;
        color: black;
        border-radius: 10px;
        outline: none; 
    }

    .document-controls button {
        display: inline-block;
        width: auto;
        margin: 5px 2px;
    }

    .additional-controls {
        display: inline-block;
        width: auto;
        margin-bottom: 15px;
    }
    
    h1 {
        font-size: 34px;
        line-height: 1.25em;
    }
    
    h2 {
        font-size: 31px;
    }

    h3 {
        font-size: 28px;
    }

    h4 {
        font-size: 25px;
    }

    h5 {
        font-size: 22px;
    }

    .metadata-header {
        font-family: monospace;
        margin-left: 12px;
        cursor: pointer; 
        position: relative; 
        white-space: pre-wrap;
        overflow-wrap: break-word;
    }

    .metadata-content {
        font-family: monospace;
        font-size: 15px;
        display: none;
        margin-left: 10px; 
        transition: max-height 0.3s ease, opacity 0.3s ease;
        overflow: hidden;
        opacity: 0;
        max-height: 0;
        margin-bottom: 10px;
    }

    .metadata.expanded .metadata-content {
        display: block;
        opacity: 1;
        max-height: 500px; 
        margin-bottom: 20px;
    }

    .ner-entity {
        cursor: default;
    }

    .sidebar {
        border-radius: 10px;
    }

    .summaries {
        background-color: #f6f6f6; 
        padding: 10px;
        border: 2px solid #e5e5e5; 
        border-radius: 10px;
    }

    #summary-content {
        font-size: 15px;
    }

    .timestamp {
        color: #f41ad7; 
        font-family: monospace; 
        font-size: 16px;
        cursor: default;
    }

    .tldr {
        margin-bottom: 15px;
        font-size: 15px;
        padding: 15px;
        border: solid 2px #e5e5e5;
        border-radius: 10px;
        cursor: pointer;
    }

    .tldr-content {
        color: brown;
        font-family: monospace;
        font-size: 15px;
    }

    .tldr-keywords {
        color: brown;
        font-family: monospace;
        font-size: 15px;
        display: none;
    }

    ol ul {
        list-style: none;
        font-size: 95%;
        line-break: loose;
        padding: 0;
        margin-top: 5px;
        margin-bottom: 5px;
        margin-left: 14px;
    }

    ul li {
        margin: 0; 
        padding: 0; 
        position: relative;
        padding-left: 10px; 
    }

    ul li::before {
        content: "•";
        position: absolute;
        left: 0; 
        margin-right: 5px;
        color: black;
        top: 0; 
    }
</style>
</head>
<body>
<div class="container">
    <div class="row">
    
        <!-- Sidebar -->
        <div class="one-third column sidebar">
            <div class="controls">
            
                <!-- Control Font-Size -->
                <div class="document-controls">
                    <button onclick="decreaseFontSize()">aa</button>
                    <button onclick="increaseFontSize()">AA</button>
                </div>

                <!-- Additional Buttons (Timestamps, NER, saving) -->
                <div class="additional-controls"></div>
                
            </div>
        </div>
        
        <div class="two-thirds column">

            <!-- Metadata -->
            <div class="metadata">
                <h6 class="metadata-header"></h6>
                <div class="metadata-content"></div>
            </div>
            
            <!-- TL;DR -->
            <div class="tldr"></div>

            <!-- Main Content -->
            <div class="content" contenteditable="true"></div>
            
        </div>
    </div>
</div>
<button onclick="scrollToTop()" id="back-to-top">Top</button>
<script>
    // Variables to keep track of the state
    var isTimestampVisible = true;
    var isHighlightingEnabled = true;
    var isSourceView = false;

    function toggleSource() {
        var contentArea = document.querySelector('.content');
        if (!isSourceView) {
            // Switch to source code view
            var htmlContent = contentArea.innerHTML;

            // Create a textarea to display the HTML code
            var textarea = document.createElement('textarea');
            textarea.value = htmlContent;
            textarea.style.width = '100%';
            textarea.style.height = '100%';
            textarea.id = 'source-editor';
            textarea.style.fontFamily = 'monospace';
            textarea.style.fontSize = '16px';
            textarea.style.lineHeight = '2';
            textarea.style.padding = '10px';
            textarea.style.border = '1px solid #ccc';
            textarea.style.backgroundColor = '#f6f6f6';
            textarea.style.borderRadius = '10px';

            // Replace the content div with the textarea
            contentArea.parentNode.replaceChild(textarea, contentArea);

            isSourceView = true;
        } else {
            // Switch back to visual view
            var textarea = document.getElementById('source-editor');
            var htmlContent = textarea.value;

            // Create a new content div
            var contentDiv = document.createElement('div');
            contentDiv.className = 'content';
            contentDiv.setAttribute('contenteditable', 'true');
            contentDiv.innerHTML = htmlContent;

            // Replace the textarea with the content div
            textarea.parentNode.replaceChild(contentDiv, textarea);

            isSourceView = false;
        }
    }

    function toggleTimestamp() {
        // Toggle the state variable
        isTimestampVisible = !isTimestampVisible;

        const timestampElements = document.querySelectorAll('.timestamp');
        timestampElements.forEach(function(element) {
            if (isTimestampVisible) {
                element.style.display = '';
            } else {
                element.style.display = 'none';
            }
        });
        var btn = document.getElementById('toggle-timestamp');
        btn.classList.toggle('active');
    }

    function toggleHighlighting() {
        // Toggle the state variable
        isHighlightingEnabled = !isHighlightingEnabled;

        const highlightedElements = document.querySelectorAll('.ner-entity');
        highlightedElements.forEach(function(element) {
            if (isHighlightingEnabled) {
                element.style.backgroundColor = element.getAttribute('data-color');
            } else {
                element.style.backgroundColor = '';
            }
        });
        var btn = document.getElementById('toggle-highlighting');
        btn.classList.toggle('active');
    }

    function changeFontSize(delta) {
        let elements = document.querySelectorAll('.content');
        elements.forEach(function(element) {
            let currentSize;

            // Try to read the current font size from the inline style
            if (element.style.fontSize) {
                currentSize = parseFloat(element.style.fontSize);
                console.log('Current font size from inline style:', currentSize, 'px');
            } else {
                // If not set, use the initial font size (e.g., 18px)
                currentSize = 18;
                console.log('No inline font size set. Using initial font size:', currentSize, 'px');
            }

            let newSize = currentSize + delta;
            console.log('New font size:', newSize, 'px');

            if (newSize >= 6 && newSize <= 48) {
                element.style.fontSize = newSize + 'px';
                console.log('Font size updated to:', newSize, 'px');
            } else {
                console.log('New font size out of bounds:', newSize, 'px');
            }
        });
    }

    function increaseFontSize() {
        changeFontSize(1); 
    }

    function decreaseFontSize() {
        changeFontSize(-1);
    }
    
    function savePage() {
        const htmlContent = document.documentElement.outerHTML;

        // Get the current page's filename
        let filename = window.location.pathname.split('/').pop();

        // If filename is empty (e.g., the URL ends with a '/'), set a default filename
        if (!filename) {
            filename = 'page';
        } else {
            // Remove any query parameters or hash fragments from the filename
            filename = filename.split('?')[0].split('#')[0];

            // Remove the file extension if it exists
            const dotIndex = filename.lastIndexOf('.');
            if (dotIndex > -1) {
                filename = filename.substring(0, dotIndex);
            }
        }

        filename = filename + '_edit.html';

        const blob = new Blob([htmlContent], { type: 'text/html' });
        const link = document.createElement('a');
        link.href = URL.createObjectURL(blob);
        link.download = filename; 

        document.body.appendChild(link);
        link.click();

        document.body.removeChild(link);
    }

    document.addEventListener('DOMContentLoaded', function() {
    // Select metadata headers
    const metadataHeaders = document.querySelectorAll('.metadata-header');

    metadataHeaders.forEach(header => {
        header.addEventListener('click', function() {
            // Toggle the 'expanded' class on the parent .metadata div
            const metadataDiv = this.parentElement;
            metadataDiv.classList.toggle('expanded');
            });
        });
    });

    // Back to Top Button
    window.onscroll = function() { scrollFunction() };

    function scrollFunction() {
        var backToTopButton = document.getElementById("back-to-top");
        if (document.body.scrollTop > 500 || document.documentElement.scrollTop > 500) {
            backToTopButton.style.display = "block";
        } else {
            backToTopButton.style.display = "none";
        }
    }

    function scrollToTop() {
        window.scrollTo({ top: 0, behavior: 'smooth' });
    }

    // Add the event listener to show/hide TL;DR content
    document.addEventListener('DOMContentLoaded', function() {
        const tldrDiv = document.querySelector('.tldr');
        tldrDiv.addEventListener('click', function() {
            const contentDiv = tldrDiv.querySelector('.tldr-content');
            const keywordsDiv = tldrDiv.querySelector('.tldr-keywords');

            // Check current display state and toggle
            if (contentDiv.style.display === 'none') {
                // If content is hidden, show it and hide keywords
                contentDiv.style.display = 'block';
                keywordsDiv.style.display = 'none';
            } else {
                // If content is shown, hide it and show keywords
                contentDiv.style.display = 'none';
                keywordsDiv.style.display = 'block';
            }
        });
    });
</script>
</body>
</html>
"""


import re
import csv
import json
from enum import Enum
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from collections import Counter
from bs4 import BeautifulSoup

from dygest import templates
from dygest.translations import UI_TRANSLATIONS


CATEGORY_COLORS = {
    'ORG': '#f5b222',
    'PER': '#7ccfff',
    'LOC': '#1bc89a',
    'DATE': '#f4eb67',
    'MISC': '#ff8222',
}


class ExportFormats(str, Enum):
    ALL = 'all'
    JSON = 'json'
    CSV = 'csv'
    HTML = 'html'


@dataclass
class WriterBase:
    filename: Path = field(default=None, init=True)
    output_filepath: Path = field(default=None, init=True)
    input_text: str = field(default=None, init=True)
    chunk_size: int = field(default=None, init=True)
    chunks: dict = field(default=None, init=True)
    named_entities: list = field(default=None, init=True)
    toc: list[dict] = field(default=None, init=True)
    summary: str = field(default=None, init=True)
    keywords: list[str] = field(default=None, init=True)
    language: str = field(default=None, init=True)
    light_model: str = field(default=None, init=True)
    expert_model: str = field(default=None, init=True)
    token_count: int = field(default=None, init=True)
    sentence_offsets: dict = field(default_factory=dict, init=True)


@dataclass
class HTMLWriter(WriterBase):
    """
    Class for writing HTML output.
    """
    export_metadata: bool = field(default=None, init=True)
    has_speaker: bool = field(default=False, init=True)
    
    def __post_init__(self):
        # HTML template → templates.py
        self.soup = BeautifulSoup(templates.HTML_CONTENT, "html.parser")
        self.has_speaker = False

    def write(self):
        """
        Builds and saves a HTML page with LLM summaries and Named Entities.
        """
        # Add metadata
        self.add_metadata_to_html()

        # Add main content
        self.add_main_content()

        # Add controls
        self.add_controls()

        # Save HTML to disk
        self.save_html()

    def format_metadata_entry(self, parent, label, value):
        span_tag = self.soup.new_tag('span')
        span_tag.string = f"{label}: {value}"
        br_tag = self.soup.new_tag('br')
        parent.append(span_tag)
        parent.append(br_tag)

    def add_metadata_to_html(self):
        """
        Adds metadata to HTML page.
        """
        div_metadata = self.soup.find('div', class_='metadata')
        if div_metadata:
            # Add filename
            h6_tag = div_metadata.find('h6', class_='metadata-header')
            if h6_tag:
                h6_tag.string = f'{self.filename}'
            
            # Add and format metadata
            if self.export_metadata:
                div_metadata_content = self.soup.find('div', class_='metadata-content')
                if div_metadata_content:
                    # Light LLM
                    self.format_metadata_entry(
                        parent=div_metadata_content, 
                        label="Light LLM (light_model)", 
                        value=self.light_model
                    )
                    # Expert LLM
                    self.format_metadata_entry(
                        parent=div_metadata_content,
                        label="Expert LLM (expert_model)", 
                        value=self.expert_model
                    )
                    # Chunk size
                    self.format_metadata_entry(
                        parent=div_metadata_content, 
                        label="Chunk size", 
                        value=self.chunk_size
                    )
                    # Datetime
                    current_time = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
                    self.format_metadata_entry(
                        parent=div_metadata_content,
                        label="Date",
                        value=current_time
                    )
                    # Token count
                    token_count = f"{self.token_count:,.0f}".replace(',', '.')
                    self.format_metadata_entry(
                        parent=div_metadata_content, 
                        label="Tokens",
                        value=token_count
                    )

                    if self.named_entities:
                        ner_count = Counter([x['ner_tag'] for x in self.named_entities])
                        ner_string = f"{len(self.named_entities)} {list(ner_count.items())}"
                        self.format_metadata_entry(
                            parent=div_metadata_content,
                            label="Entities",
                            value=ner_string
                        )

    def add_controls(self):
        # Append additional button controls to HTML
        div_additional_controls = self.soup.find('div', class_='additional-controls')
        
        if div_additional_controls:
            # Add NER Highlighting Button  
            if self.named_entities:
                button_NER_highlighting = self.soup.new_tag(
                    'button',
                    attrs={
                        "id": "toggle-highlighting",
                        "onclick": "toggleHighlighting()"
                    })
                button_NER_highlighting.string = get_translation(
                    key='button_NER_highlighting',
                    language_code=self.language
                )
                div_additional_controls.append(button_NER_highlighting)

            # Add Speaker / Timestamp Button 
            if self.has_speaker:
                button_timestamps = self.soup.new_tag(
                    'button',
                    attrs={
                        "id": "toggle-timestamp",
                        "onclick": "toggleTimestamp()"
                    })
                button_timestamps.string = get_translation(
                    key='button_timestamps',
                    language_code=self.language
                )
                div_additional_controls.append(button_timestamps)

            # Add button for showing HTML Code
            button_show_html = self.soup.new_tag(
                'button',
                attrs={
                    'id': 'toggle-source',
                    'onclick': 'toggleSource()'
                })
            button_show_html.string = get_translation(
                key='button_show_HTML',
                language_code=self.language
                )
            div_additional_controls.append(button_show_html)

        # Add button for saving the edited HTML
        div_document_controls = self.soup.find('div', class_='document-controls')
        button_save = self.soup.new_tag(
            'button',
            attrs={
                'class': 'save',
                'onclick': 'savePage()'
            })
        button_save.string = get_translation(
                key='button_save',
                language_code=self.language
            )
        div_document_controls.append(button_save)

    def add_main_content(self):
        """
        Appends main content blocks to a default HTML template.
        """
        # Add summary / TL;DR
        if self.summary:
            div_tldr = self.soup.find('div', class_='tldr')
            if div_tldr:
                # Create new div element
                div_summary = self.soup.new_tag(
                    'div',
                    attrs={'class': 'tldr-content'}
                )

                # Add title tag
                title_tag = self.soup.new_tag('span')
                title_tag.string = get_translation(
                    key="summary",
                    language_code=self.language
                )       

                # Add summary
                summary_tag = self.soup.new_tag('span')
                summary_tag.string = self.summary.strip()

                # Append new tags
                div_summary.append(title_tag)
                div_summary.append(summary_tag)
                div_tldr.append(div_summary)
        else:
            div_tldr = self.soup.find('div', class_='tldr')
            if div_tldr:
                # Create new div element
                div_summary = self.soup.new_tag(
                    'div',
                    attrs={'class': 'tldr-content'}
                )

                # Add title tag
                title_tag = self.soup.new_tag('span')
                title_tag.string = "No Summary!"
                
                # Append new tags
                div_summary.append(title_tag)
                div_tldr.append(div_summary)
                
        # Add keywords
        if self.keywords:
            div_tldr = self.soup.find('div', class_='tldr')
            if div_tldr:
                # Create new div element
                div_keywords = self.soup.new_tag(
                    'div',
                    attrs={'class': 'tldr-keywords'}
                )

                # Add title tag
                title_tag = self.soup.new_tag('span')
                title_tag.string = get_translation(
                    key="keywords",
                    language_code=self.language
                )       

                # Add keywords string
                keywords_tag = self.soup.new_tag('span')
                keywords_tag.string = (
                    ', '.join(
                        [key if key.isupper() else key.title() for key in self.keywords]
                        )
                    )

                # Append new tags
                div_keywords.append(title_tag)
                div_keywords.append(keywords_tag)
                div_tldr.append(div_keywords)
        
        # Add TOC heading
        if self.toc:
            div_additional_controls = self.soup.find(
                'div', 
                class_='additional-controls'
                )
            div_summaries = self.soup.new_tag(
                'div', 
                attrs={'class': 'summaries'}
                )
            
            div_summaries_content = self.soup.new_tag(
                'div',
                attrs={
                    'id': 'summary-content',
                    'style': 'display: block; text-align: left'
                }
            )
            
            h5_tag = self.soup.new_tag(
                'h5',
                attrs={
                    'style': 'text-align: center;'
                }
            )
            h5_tag.string = get_translation(
                key='heading_topics',
                language_code=self.language
            )
            div_summaries_content.append(h5_tag)
            
            ol_tag = self.soup.new_tag('ol')
            div_summaries_content.append(ol_tag)
            div_summaries.append(div_summaries_content)
            div_additional_controls.insert_after(div_summaries)

            # Add TOC topics and links
            ol_tag = self.soup.find("ol")

            for item in self.toc:
                # Create a list element for the headline
                li_tag = self.soup.new_tag("li")
                strong_tag = self.soup.new_tag("strong")
                strong_tag.string = f'{item["headline"]}'
                li_tag.append(strong_tag)
                
                # Create an unordered list for topics
                ul_tag = self.soup.new_tag("ul")

                # Append each topic to the unordered list
                for topic in item['topics']:
                    topic_li = self.soup.new_tag('li')
                    
                    # Get location to create a link to the content
                    location = topic.get('location')
                    if location:
                        link_id = f"location-{location}"
                        link_tag = self.soup.new_tag("a", href=f"#{link_id}")
                        link_tag.string = f"{topic['summary'].rstrip('.,;:?!')}"
                        topic_li.append(link_tag)
                    
                    # Append topics to unordered list
                    ul_tag.append(topic_li)
                
                li_tag.append(ul_tag)
                ol_tag.append(li_tag)


        ## Main Content
        
        events = []
        
        ### 1. Collect NER events
        if self.named_entities:
            for entity in self.named_entities:
                entity_start = entity['start']
                entity_end = entity['end']
                entity_text = entity['text']
                entity_ner_tag = entity['ner_tag']

                # Skip certain entities
                if entity_text in ['SPEAKER', 'UNKNOWN', '.']:
                    continue

                events.append((entity_start, 'start_entity', entity_ner_tag))
                events.append((entity_end, 'end_entity', None))
        
        ### 2. Match timestamp / speaker patterns
        pattern = r'\[\d{2}:\d{2}:\d{2}\.\d{3}\] \[(?:SPEAKER_\d+|UNKNOWN)\]|\[(?:SPEAKER_\d+|UNKNOWN)\]'
        
        for match in re.finditer(pattern, self.input_text):
            start, end = match.start(), match.end()
            events.append((start, 'insert_linebreak', None))
            events.append((start, 'start_speaker', None))
            events.append((end, 'end_speaker', None))
            self.has_speaker = True  
        
        ### 3. Collect anchor tag events for TOC (now using sentence_offsets)
        inserted_locations = set()
        if self.toc:
            for item in self.toc:
                for topic in item['topics']:
                    location = topic.get('location')
                    if location and location not in inserted_locations:
                        # Check if location (e.g. "S12") in sentence_offsets
                        if location in self.sentence_offsets:
                            start = self.sentence_offsets[location]
                            anchor_id = f"location-{location}"
                            events.append((start, 'insert_anchor', anchor_id))
                            inserted_locations.add(location)
                            
                            # Do not insert a linebreak if speaker tag was found
                            if self.has_speaker:
                                continue
                            
                            events.append((start, 'insert_linebreak', None))
        
        ### 4. Insert linebreak events for each "\n"
        for match in re.finditer(r'\n{2}', self.input_text):
            events.append((match.start(), 'insert_linebreak', None))
            
                        
        # Define event order to handle conflicts
        event_order = {
            'end_entity': 0,
            'end_speaker': 1,
            'insert_anchor': 2,
            'insert_linebreak': 3,
            'start_speaker': 4,
            'start_entity': 5
        }

        events.sort(key=lambda x: (x[0], event_order[x[1]]))
        
        # Build the content with paragraphs
        p_tag = self.soup.new_tag('p')
        content_elements = []
        current_parent = p_tag
        spans_stack = []
        position = 0
        
        for event in events:
            event_position, event_type, data = event
            
            # Add text up to the event position
            if position < event_position:
                text_segment = self.input_text[position:event_position]
                current_parent.append(text_segment)
            position = event_position
            
            if event_type == 'start_entity':
                # Save the current parent to the stack
                spans_stack.append(current_parent)

                # Create the outer span for the entity
                outer_span_tag = self.soup.new_tag(
                    "span",
                    attrs={
                        "class": "ner-entity",
                        "data-color": CATEGORY_COLORS.get(data, ''),
                        "style": f"background-color: {CATEGORY_COLORS.get(data, '')};",
                        "title": data,
                        "contenteditable": False
                    })

                # Create the inner span that will hold the text
                inner_span_tag = self.soup.new_tag(
                    "span",
                    attrs={
                        "class": "edits",
                        "tabindex": 0,
                        "contenteditable": True
                    })

                outer_span_tag.append(inner_span_tag)
                current_parent.append(outer_span_tag)

                # Set the current parent to the inner span
                current_parent = inner_span_tag
                
            elif event_type == 'end_entity':
                # Close the current entity span
                if spans_stack:
                    spans_stack.pop()
                    current_parent = spans_stack[-1] if spans_stack else p_tag
                else:
                    print(f'span_stack pop Error for event: {event}')
                    current_parent = spans_stack[-1] if spans_stack else p_tag
                    
            elif event_type == 'insert_anchor':
                # Ensure all spans are closed so the anchor doesn't appear 
                # nested incorrectly. If necessary, close spans:
                spans_stack.clear()
                current_parent = p_tag

                # Create and insert the anchor inline
                anchor_tag = self.soup.new_tag(
                    "a",
                    id=data,
                    attrs={
                        "class": "anchor"
                    }
                )
                current_parent.append(anchor_tag)
            
            elif event_type == 'insert_linebreak':
                # Close any open spans before ending the paragraph
                spans_stack.clear()
                current_parent = p_tag
                
                # Add the current paragraph to the content if it has any children
                if p_tag.contents:
                    content_elements.append(p_tag)
                
                # Start a new paragraph
                p_tag = self.soup.new_tag('p')
                current_parent = p_tag
            
            elif event_type == 'start_speaker':
                spans_stack.append(current_parent)
                
                outer_span_tag = self.soup.new_tag(
                    "span", 
                    attrs={
                        "class": "timestamp",
                        "contenteditable": False
                    })
                
                inner_span_tag = self.soup.new_tag(
                    "span",
                    attrs={
                        "class": "edits",
                        "tabindex": 0,
                        "contenteditable": True
                    })
                
                outer_span_tag.append(inner_span_tag)
                current_parent.append(outer_span_tag)
                current_parent = inner_span_tag
            
            elif event_type == 'end_speaker':
                # Close the speaker span
                spans_stack.pop()
                current_parent = spans_stack[-1] if spans_stack else p_tag
        
        # Add any remaining text after the last event
        if position < len(self.input_text):
            text_segment = self.input_text[position:]
            current_parent.append(text_segment)
        
        # Add the last paragraph to the content if it has any children
        if p_tag.contents:
            content_elements.append(p_tag)
        
        # Write the new content
        content_div = self.soup.find('div', class_='content')
        for element in content_elements:
            content_div.append(element)

        # Transform any markdown tags to HTML
        self.transform_markdown_tags()
        
    def transform_markdown_tags(self):
        """
        Transforms markdown syntax within the main content of the HTML template
        to corresponding HTML tags.
        """
        # Locate the main content div
        main_content = self.soup.find('div', class_='content')
        if not main_content:
            print(f"""... No <div class="content"> in HTML.""")
            return

        markdown_patterns = {
            'horizontal_rule': re.compile(r'^-{3,}$', re.MULTILINE),
            'blockquotes': re.compile(r'^>\s?(.*)', re.MULTILINE),
            'headings': re.compile(r'^(#{1,6})\s*(.+)', re.MULTILINE),
            'ordered_list': re.compile(r'^(\s*)\d+\.\s+(.*)', re.MULTILINE),
            'unordered_list': re.compile(r'^(\s*)[-\*\+]\s+(.*)', re.MULTILINE),
            'images': re.compile(r'!\[([^\]]*)\]\(([^\)]+)\)'),
            'links': re.compile(r'\[([^\]]+)\]\(([^\)]+)\)'),
            'inline_code': re.compile(r'`([^`]*)`'),
            'bold': re.compile(r'(\*\*|__)(.*?)\1'),
            'italics': re.compile(r'(\*|_)(.*?)\1'),
            'strikethrough': re.compile(r'~~(.*?)~~'),
        }

        def replace_heading(match):
            hashes, text = match.groups()
            level = len(hashes)
            return f'<h{level}>{text.strip()}</h{level}>'

        def replace_ordered_list(match):
            indent, item = match.groups()
            return f'{indent}<ol><li>{item.strip()}</li></ol>'

        def replace_unordered_list(match):
            indent, item = match.groups()
            return f'{indent}<ul><li>{item.strip()}</li></ul>'

        # Convert the main_content to a string for regex operations
        content_html = str(main_content)

        # Apply horizontal rule transformation first to avoid conflicts
        content_html = markdown_patterns['horizontal_rule'].sub(
            '<hr/>', 
            content_html
            )

        # Blockquotes
        content_html = markdown_patterns['blockquotes'].sub(
            r'<blockquote>\1</blockquote>', 
            content_html
            )

        # Headings
        content_html = markdown_patterns['headings'].sub(
            replace_heading, 
            content_html
            )

        # Lists
        content_html = markdown_patterns['ordered_list'].sub(
            replace_ordered_list, 
            content_html
            )
        
        content_html = markdown_patterns['unordered_list'].sub(
            replace_unordered_list, 
            content_html
            )

        # Images
        content_html = markdown_patterns['images'].sub(
            r'<img alt="\1" src="\2" />', 
            content_html
            )

        # Links
        content_html = markdown_patterns['links'].sub(
            r'<a href="\2">\1</a>', 
            content_html
            )

        # Inline code
        content_html = markdown_patterns['inline_code'].sub(
            r'<code>\1</code>', 
            content_html
            )

        # Bold text
        content_html = markdown_patterns['bold'].sub(
            r'<strong>\2</strong>', 
            content_html
            )

        # Italics
        content_html = markdown_patterns['italics'].sub(
            r'<em>\2</em>', 
            content_html
            )

        # Strikethrough
        content_html = markdown_patterns['strikethrough'].sub(
            r'<del>\1</del>', 
            content_html
            )

        # Parse the transformed HTML with BeautifulSoup
        transformed_soup = BeautifulSoup(content_html, 'html.parser')
        
        # Clean empty <p> tags
        for p in transformed_soup.find_all('p'):
            # Check if the paragraph has neither text nor any child elements
            if not p.get_text(strip=True) and not p.find(True, recursive=False):
                p.decompose()

        # Replace the original main_content with the transformed content
        main_content.clear()
        for element in transformed_soup.contents:
            main_content.append(element)

        self.soup.prettify(formatter='html')

    def save_html(self):
        with open(self.output_filepath, 'w', encoding='utf-8') as fout:
            fout.write(str(self.soup))
        print(f"... 🌞 Saved HTML → {self.output_filepath}")


@dataclass
class CSVWriter(WriterBase):
    """
    Class for writing CSV output.
    """
    def write(self):
        with open(self.output_filepath, 'w', encoding='utf-8') as fout:
            csv_header = [
                'filename',
                'output_filepath',
                'language',
                'chunk_size',
                'token_count',
                'light_model',
                'expert_model',
                'chunks',
                'summary' if self.summary else None,
                'keywords' if self.keywords else None,
                'toc' if self.toc else None
            ]
            writer = csv.writer(fout)
            writer.writerow(csv_header)
            writer.writerow(
                [
                    self.filename,
                    self.output_filepath,
                    self.language,
                    self.chunk_size,
                    self.token_count,
                    self.light_model,
                    self.expert_model,
                    self.chunks,
                    self.summary if self.summary else None,
                    self.keywords if self.keywords else None,
                    self.toc if self.toc else None
                ]
            )
        print(f"... 🌞 Saved CSV → {self.output_filepath}")


@dataclass
class JSONWriter(WriterBase):
    """
    Class for writing JSON output.
    """    
    def write(self):
        with open(str(self.output_filepath), 'w', encoding='utf-8') as fout:
            json_dict = {
                'filename': self.filename,
                'output_filepath': str(self.output_filepath),
                'language': self.language,
                'chunk_size': self.chunk_size,
                'token_count': self.token_count,
                'light_model': self.light_model,
                'expert_model': self.expert_model,
                'chunks': self.chunks,
                'summary': self.summary if self.summary else None,
                'keywords': self.keywords if self.keywords else None,
                'toc': self.toc if self.toc else None
            }
            json.dump(json_dict, fout, indent=4)
        print(f"... 🌞 Saved JSON → {self.output_filepath}")
        

def get_writer(proc):
    """
    Instantiates and returns the appropriate writer Class based on 
    proc.export_format, including shared and format-specific parameters.
    """
    def html_specific_params(proc):
        return {'export_metadata': proc.export_metadata} if proc.export_metadata else {}
    
    def csv_specific_params(proc):
        return {}
    
    def json_specific_params(proc):
        return {}
        
    # Map ExportFormats to (WriterClass, specific_params_function)
    writer_mapping = {
        ExportFormats.HTML: (HTMLWriter, html_specific_params),
        ExportFormats.CSV: (CSVWriter, csv_specific_params),
        ExportFormats.JSON: (JSONWriter, json_specific_params)
    }

    # Retrieve writer_class and specific params function based on export_format
    writer_entry = writer_mapping.get(proc.export_format)
    if not writer_entry:
        raise ValueError(f"... Unknown export format: {proc.export_format}")

    writer_class, specific_params_func = writer_entry

    # Shared params for all writer classes
    shared_params = {
        'filename': proc.filename,
        'input_text': proc.text,
        'chunk_size': proc.chunk_size,
        'chunks': proc.chunks,
        'named_entities': proc.entities if proc.add_ner else None,
        'toc': proc.toc if proc.add_toc else None,
        'summary': proc.summaries if proc.add_summaries else None,
        'keywords': proc.keywords if proc.add_keywords else None,
        'language': proc.language_ISO,
        'light_model': proc.light_model,
        'expert_model': proc.expert_model,
        'token_count': proc.token_count,
        'sentence_offsets': proc.sentence_offsets
    }

    # Get file suffix based on export_format
    suffix = proc.export_format.value.lower()
    output_filepath = proc.output_filepath.with_suffix(f'.{suffix}')
    shared_params['output_filepath'] = output_filepath

    # Get specific params and merge them
    specific_params = specific_params_func(proc)
    specific_params = {k: v for k, v in specific_params.items() if v is not None}
    writer_params = {**shared_params, **specific_params}

    # Return the writer
    return writer_class(**writer_params)

def get_translation(key, language_code, default_language='en'):
    translations = UI_TRANSLATIONS.get(key, {})
    return translations.get(
        language_code, translations.get(default_language, '')
        )


import re
import os
import time
import typer
import warnings
from rich import print
from typing import Optional
from openai import OpenAIError
from litellm import (
    completion,
    embedding,
    supports_response_schema,
    BadRequestError
    )

def get_api_base(model_name: str) -> Optional[str]:
    """
    Determine the api_base URL based on the LLM service provider.
    """
    api_base_mappings = {
        'ollama': 'http://localhost:11434'
    }
    provider = model_name.split('/')[0].lower()
    return api_base_mappings.get(provider, None)

def get_api_key(model_name: str) -> Optional[str]:
    """
    Determine and get the api_key based on the LLM service provider.
    """
    provider = model_name.split('/')[0].lower()
    if provider != 'ollama':
        api_key = os.getenv(f'{provider.upper()}_API_KEY')
        if not api_key:
            print(f"Please set your {provider.upper()}_API_KEY in your \
environment using: export {provider.upper()}_API_KEY='your_api_key'")
        return api_key
    return None

def set_llm_response_format(
        model: str,
        output_format: str, 
        json_schema: dict
    ) -> dict:
    """
    Construct the response format configuration for an LLM call based on the 
    desired output type and JSON schema.

    The function supports two output formats:
      - 'text': Returns a simple dictionary specifying plain text output.
      - 'json_schema': Checks if the current model supports structured output
        by using the supports_response_schema function. If supported:
          - If no JSON schema is provided, returns a basic 'json_schema' type.
          - If a JSON schema is provided and the provider is 'groq', returns
            the schema under the key 'json_object' (to accommodate groq API
            requirements).
          - Otherwise, returns the schema under the key 'json_schema'.
        If structured output is not supported, it falls back to plain text 
        output.

    Parameters:
        model (str): The LLM model identifier formatted as 'provider/model'.
        output_format (str): Desired output format: 'text' or 'json_schema'.
        json_schema (dict): A dictionary defining the expected JSON schema 
        when structured output is desired.

    Returns:
        dict: A dictionary that specifies the response format configuration 
        for the LLM API call.
    """
    if output_format == 'text':
        response_format = {
            'type': 'text'
            }
    elif output_format == 'json_schema':
        # Check if current model supports structured output
        provider, *_, llm_model = model.split('/')
        if supports_response_schema(llm_model, provider):
            if not json_schema:
                response_format = {
                    'type': 'json_schema'
                    }
            elif json_schema and model.split('/')[0] == 'groq':
                response_format = {
                    'type': 'json_schema',
                    'json_object': json_schema # groq API uses different key
                }
            else:
                response_format = {
                    'type': 'json_schema',
                    'json_schema': json_schema
                } 
        else:
            response_format = {
                'type': 'text'
            }

    return response_format

def remove_reasoning(llm_response: str = None) -> str:
    """
    Clean <think></think> parts from reasoning models responses.
    """
    pattern = r"<think>.*?</think>"
    return re.sub(
        pattern, 
        '', 
        llm_response.choices[0].message.content,
        flags=re.DOTALL
        )

def call_llm(
    prompt: str = None, 
    model: str = None,
    output_format: str = 'text', # 'json_schema' or 'text'
    json_schema: None | dict = None,
    temperature: float = 0.1, 
    api_key: str = None,
    api_base: str = None,
    sleep_time: float = 0.0
    ):
    """
    Call an LLM using litellm's completion function.
    Note: Set the environment variable for the chosen provider's API key.
    
    Parameters:
    - prompt: The prompt template
    - text_input: The input text to be summarized or processed
    - model: The LLM model (e.g., 'openai/gpt-3.5-turbo', 'anthropic/claude-2',
             'huggingface/...', 'ollama/...') 
    - temperature: The LLM sampling temperature
    - api_base: Optional API base URL for some models (like Ollama or custom 
                Hugging Face endpoints)
    """
    # Check for api_key and api_base
    if not api_base or api_key:
        api_base = get_api_base(model)
        api_key = get_api_key(model)

    # Set LLM response output formats ('json_schema' or 'text')
    response_format = set_llm_response_format(model, output_format, json_schema)

    try:
        # Catch httpx deprecation warning
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", 
                category=DeprecationWarning, 
                module="httpx._content"
                )
            response = completion(
                model=model,
                messages=[{
                    "role": "user",
                    "content": prompt
                    }],
                temperature=temperature,
                api_key=api_key,
                base_url=api_base,
                response_format=response_format
                )

        if sleep_time > 0:
            time.sleep(sleep_time)
        
        # Clean <think> responses from reasoning models like DeepSeek
        response = remove_reasoning(response)

        return response

    except BadRequestError as e:
        print(f"[purple] ...  Error: Please configure the model(s) your are \
using with the correct LLM provider (e.g. 'ollama/llama3.1:latest', \
'openai/gpt-4o-mini).")
        raise typer.Exit(code=1) 
    except OpenAIError as e:
        print(f"[purple] ... Error: {e}")
        raise typer.Exit(code=1) 
    except Exception as e:
        print(f"[purple] ... An unexpected error occurred: {e}")
        raise typer.Exit(code=1) 

def get_embeddings(
    text: str, 
    model: str,
    api_key: str = None,
    api_base: str = None
    ):
    """
    Retrieve embeddings from a given model using litellm.embed.
    
    Parameters:
    - text: The text to embed
    - model: The embedding model (e.g., 'openai/text-embedding-ada-002',
            'huggingface/...')
    - api_base: Optional API base URL for custom endpoints
    
    Returns:
    A dictionary containing the embeddings.
    """
    if not api_base or api_key:
        api_base = get_api_base(model)
        api_key = get_api_key(model)
    
    try:
        response = embedding(
            input=text,
            model=model,
            api_key=api_key,
            api_base=api_base
        )
        
        return response.data[0]['embedding']
    
    except OpenAIError as e:
        print(f"[purple] ... An OpenAI API error occurred: {e}")
        raise typer.Exit(code=1) 
    except Exception as e:
        print(f"[purple] ... An unexpected error occurred: {e}")
        raise typer.Exit(code=1) 



import yaml
import typer
from rich import print
from enum import Enum
from pathlib import Path

DEFAULT_CONFIG = {
    'light_model': None,
    'expert_model': None,
    'embedding_model': None,
    'temperature': 0.1,
    'api_base': None,
    'chunk_size': 1000,
    'ner': False,
    'language': 'auto',
    'precise': False,
    'sleep': 2.5
}

CONFIG_FILE = Path.cwd() / "dygest_config.yaml"
CONFIG = None

def missing_config_requirements(config: dict) -> bool:
    """
    Checks if all required CONFIG fields are set by the user.
    """
    required_fields = [
        'light_model', 
        'expert_model', 
        'embedding_model'
        ]
    return any(config.get(field) is None for field in required_fields)

def load_config() -> dict:
    """
    Loads the configuration from a JSON file if it exists.
    Otherwise, returns the default configuration.
    Converts the 'language' field to the NERlanguages Enum if possible.
    """
    from dygest.ner_utils import NERlanguages

    if CONFIG_FILE.exists():
        with open(CONFIG_FILE, 'r', encoding='utf-8') as f:
            try:
                config = yaml.safe_load(f)
            except yaml.YAMLError as e:
                print(f"[purple]... Error parsing YAML configuration: {e}", err=True)
                
            # Merge with DEFAULT_CONFIG to ensure all keys are present
            merged_config = DEFAULT_CONFIG.copy()
            merged_config.update(config)

            # Convert 'language' from string to Enum if possible
            if 'language' in merged_config and isinstance(merged_config['language'], str):
                try:
                    merged_config['language'] = NERlanguages(merged_config['language'])
                except ValueError:
                    print(f"[purple]... Warning: '{merged_config['language']}' is not a valid NER language. Defaulting to 'auto'.", err=True)
                    merged_config['language'] = NERlanguages.AUTO
            else:
                merged_config['language'] = NERlanguages.AUTO

            return merged_config
            
    else:
        config = DEFAULT_CONFIG.copy()
        config['language'] = 'auto'
        return config

def save_config(config: dict):
    """
    Saves the configuration to a JSON file.
    Converts Enum fields to their string values for JSON compatibility.
    """
    config_to_save = config.copy()
    if isinstance(config_to_save.get('language'), Enum):
        config_to_save['language'] = config_to_save['language'].value

    try:
        with open(CONFIG_FILE, 'w', encoding='utf-8') as f:
            yaml.safe_dump(config_to_save, f, default_flow_style=False, sort_keys=False)
    except yaml.YAMLError as e:
        typer.echo(f"Error writing YAML configuration: {e}", err=True)

def print_config(config: dict):
    """
    Prints the LLM and embedding configuration in a structured format.
    """
    # Check if requrired fields are set
    if missing_config_requirements(config):
        print("[purple]... Please configure all required fields.")
    
    language_val = config.get('language')
    if isinstance(language_val, Enum):
        language_val = language_val.value  # If Enum, get its value
    elif not language_val:
        language_val = 'auto'

    formatted_config = {
        "Light LLM (required)": config.get('light_model') or "None",
        "Expert LLM (required)": config.get('expert_model') or "None",
        "Embedding model (required)": config.get('embedding_model') or "None",
        "Temperature": config.get('temperature', 0.0),
        "Sleep": f"{config.get('sleep', 0.0)} second(s)",
        "Chunk size": config.get('chunk_size', 0),
        "NER": config.get('ner', False),
        "NER precise": config.get('precise', False),
        "NER language": language_val,
        "API Base": config.get('api_base') or "None",
    }
    
    for key, value in formatted_config.items():
        print(f"[deep_pink2]... ... {key:<27}→ [light_pink4]{value}")



### JSON SCHEMAS ###

GET_TOPICS_JSON = {
    "name": "get_topics",
    "schema": {
        "type": "object",
        "properties": {
            "topics": {
                "type": "array",
                "items": {
                    "topic": {
                        "type": "string",
                        "description": "Concise Subheading in {language}"
                    },
                    "summary": {
                        "type": "string",
                        "description": "Concise sub-headline in {language}"
                    },
                    "location": {
                        "type": "string",
                        "description": "S<number>"
                    }
                },
                "required": ["topic", "summary", "location"]
            }
        }
    }
}

CREATE_TOC_JSON = {
    "name": "create_toc",
    "schema": {
        "type": "object",
        "properties": {
            "headline": {
                "type": "string",
                "description": "The main headline for the table of contents"
            },
            "topics": {
                "type": "object",
                "properties": {
                    "summary": {
                        "type": "string",
                        "description": "A brief summary of the topic"
                    },
                    "location": {
                        "type": "string",
                        "description": "location"
                    }
                },
                "required": ["summary", "location"]
            }
        },
        "required": ["headline", "topics"]
    }
}

{
  "summary": "Summary",
  "keywords": ["keyword1", "keyword2", ...]
}


CREATE_SUMMARY_AND_KEYWORDS_JSON = {
    "name": "create_summary_and_keywords",
    "schema": {
        "type": "object",
        "properties": {
            "summary": {
                "type": "Summary",
                "description": "The main headline for the table of contents"
            },
            "keywords": {
                "type": "array",
                "items": {
                    "keyword": {
                        "type": "string",
                        "description": "A descriptive keyword"
                    },

                },
                "required": ["summary", "keywords"]
            }
        },
        "required": ["summary", "keywords"]
    }
}


UI_TRANSLATIONS = {
    'button_NER_highlighting': {
        'ar': 'التوكيد',
        'en': 'Highlight',
        'da': 'Fremhævelse',
        'de': 'Hervorheben',
        'fr': 'Marquer',
        'es': 'Marca',
        'nl': 'Benadrukken'
    },
    'button_timestamps': {
        'ar': 'الطابع الزمني',
        'en': 'Timestamps',
        'da': 'Tidsstempel',
        'de': 'Zeitstempel',
        'fr': 'Horodatage',
        'es': 'Marca de tiempo',
        'nl': 'Tijdstempel'
    },
    'button_show_HTML': {
        'ar': 'إظهار HTML',
        'en': 'Show HTML',
        'da': 'Vis HTML',
        'de': 'HTML anzeigen',
        'fr': 'Afficher HTML',
        'es': 'Mostrar HTML',
        'nl': 'HTML weergeven'
    },
    'button_save': {
        'ar': 'الحفظ',
        'en': 'Save',
        'da': 'Gemme',
        'de': 'Speichern',
        'fr': 'Enregistrer',
        'es': 'Guardar',
        'nl': 'Sla'
    },
    'heading_topics': {
        'ar': 'المواضيع',
        'en': 'Topics',
        'da': 'Temaer',
        'de': 'Themen',
        'fr': 'Thèmes',
        'es': 'Temas',
        'nl': 'Onderwerpen'
    },
    'summary': {
        'ar': 'الملخص: ',
        'en': 'Summary: ',
        'da': 'Sammenfatning: ',
        'de': 'Zusammenfassung: ',
        'fr': 'Résumé: ',
        'es': 'Resumen: ',
        'nl': 'Samenvatting: '
    },
    'keywords': {
        'ar': 'الكلمات المفتاحية: ',
        'en': 'Keywords: ',
        'da': 'Nøgleord: ',
        'de': 'Schlagworte: ',
        'fr': 'Mots-clés: ',
        'es': 'Palabras clave: ',
        'nl': 'Trefwoorden: '
    }
}


LANGUAGES = {
    "en": "english",
    "zh": "chinese",
    "de": "german",
    "es": "spanish",
    "ru": "russian",
    "ko": "korean",
    "fr": "french",
    "ja": "japanese",
    "pt": "portuguese",
    "tr": "turkish",
    "pl": "polish",
    "ca": "catalan",
    "nl": "dutch",
    "ar": "arabic",
    "sv": "swedish",
    "it": "italian",
    "id": "indonesian",
    "hi": "hindi",
    "fi": "finnish",
    "vi": "vietnamese",
    "he": "hebrew",
    "uk": "ukrainian",
    "el": "greek",
    "ms": "malay",
    "cs": "czech",
    "ro": "romanian",
    "da": "danish",
    "hu": "hungarian",
    "ta": "tamil",
    "no": "norwegian",
    "th": "thai",
    "ur": "urdu",
    "hr": "croatian",
    "bg": "bulgarian",
    "lt": "lithuanian",
    "la": "latin",
    "mi": "maori",
    "ml": "malayalam",
    "cy": "welsh",
    "sk": "slovak",
    "te": "telugu",
    "fa": "persian",
    "lv": "latvian",
    "bn": "bengali",
    "sr": "serbian",
    "az": "azerbaijani",
    "sl": "slovenian",
    "kn": "kannada",
    "et": "estonian",
    "mk": "macedonian",
    "br": "breton",
    "eu": "basque",
    "is": "icelandic",
    "hy": "armenian",
    "ne": "nepali",
    "mn": "mongolian",
    "bs": "bosnian",
    "kk": "kazakh",
    "sq": "albanian",
    "sw": "swahili",
    "gl": "galician",
    "mr": "marathi",
    "pa": "punjabi",
    "si": "sinhala",
    "km": "khmer",
    "sn": "shona",
    "yo": "yoruba",
    "so": "somali",
    "af": "afrikaans",
    "oc": "occitan",
    "ka": "georgian",
    "be": "belarusian",
    "tg": "tajik",
    "sd": "sindhi",
    "gu": "gujarati",
    "am": "amharic",
    "yi": "yiddish",
    "lo": "lao",
    "uz": "uzbek",
    "fo": "faroese",
    "ht": "haitian creole",
    "ps": "pashto",
    "tk": "turkmen",
    "nn": "nynorsk",
    "mt": "maltese",
    "sa": "sanskrit",
    "lb": "luxembourgish",
    "my": "myanmar",
    "bo": "tibetan",
    "tl": "tagalog",
    "mg": "malagasy",
    "as": "assamese",
    "tt": "tatar",
    "haw": "hawaiian",
    "ln": "lingala",
    "ha": "hausa",
    "ba": "bashkir",
    "jw": "javanese",
    "su": "sundanese",
    "yue": "cantonese",
}
